// DCC Tender Locomotive Sound Decoder with sound for DS-DCC decode
// By yaasan & Fujigaya2
// Based on Nicolas's sketch http://blog.nicolas.cx
// Inspired by Geoff Bunza and his 17 Function DCC Decoder & updated library
//
//
// Debug serial output available on the serial port at baud 115200, aka Tools -> Serial Monitor
//

#include <NmraDcc.h>
#include <stdint.h>
#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>	 //required by notifyCVRead() function if enabled below

#include "blast_sound.h"
#include "draft_sound1_100.h"
#include "draft_sound2_100.h"
#include "kiteki_short.h"
#include "stop_sound1_150.h"
#include "stop_sound2_150.h"



/* 各種設定、宣言 */

#define DECODER_ADDRESS 3
#define DCC_ACK_PIN 9	//if defined enables the ACK pin functionality. Comment out to disable.

#define MOTOR_PWM_A 9
#define MOTOR_PWM_B 10
#define MOTOR_CVACK_POWER 75


#define MAX_PWMDUTY 200
#define MID_PWMDUTY 100

#define MOTOR_LIM_MAX 255

#define CV_VSTART		2
#define CV_ACCRATIO		3
#define CV_DECCRATIO	4

/* Class, 変数宣言  */


//使用クラスの宣言
NmraDcc	 Dcc;
DCC_MSG	 Packet;

//Task Schedule
unsigned long gPreviousL5 = 0;

//進行方向
uint8_t gDirection = 128;

//Function State
uint8_t gState_F0 = 0;
uint8_t gState_F1 = 0;
uint8_t gState_F2 = 0;
uint8_t gState_F3 = 0;
uint8_t gState_F4 = 0;

//モータ制御関連の変数
uint32_t gSpeedRef = 1;

//モータ制御関連の変数
uint32_t gPwmLPF_buf = 0;
uint32_t gPwmLPF2_buf = 0;
uint8_t gPwmDir = 128;
uint16_t gSpeedCmd = 0;
uint16_t gPrevSpeed = 0;
uint8_t gCVAckDir = 0;


//CV related
uint8_t gCV1_SAddr = 3;
uint8_t gCVx_LAddr = 3;
uint8_t gCV2_Vstart = 32;
uint8_t gCV3_AccRatio = 32;
uint8_t gCV4_DecRatio = 32;


//Sound played flag
uint8_t gSound_played = 0;
uint8_t gSound_playnext = 0;
int gSound_Length = 0;

//Sound Frequency
uint32_t sample_rate = 8000;
//Sound_delay_offset_time
int delay_offset_time = 0;

//Sound
enum SoundState {  s_start, s_sound , s_end,s_idle};
SoundState s_state = s_idle;


int ledPin = 13;
int speakerPin = 3; // Can be either 3 or 11, two PWM outputs connected to Timer 2
volatile uint16_t sample;
byte lastSample;
byte firstSample;
int repeat_sound;

//Internal variables and other.
#if defined(DCC_ACK_PIN)
const int DccAckPin = DCC_ACK_PIN ;
#endif


struct CVPair{
  uint16_t	CV;
  uint8_t	Value;
};
CVPair FactoryDefaultCVs [] = {
  {CV_MULTIFUNCTION_PRIMARY_ADDRESS, DECODER_ADDRESS},
  {CV_ACCESSORY_DECODER_ADDRESS_MSB, 0},		//The LSB is set CV 1 in the libraries .h file, which is the regular address location, so by setting the MSB to 0 we tell the library to use the same address as the primary address. 0 DECODER_ADDRESS
  {CV_MULTIFUNCTION_EXTENDED_ADDRESS_MSB, 0},	 //XX in the XXYY address
  {CV_MULTIFUNCTION_EXTENDED_ADDRESS_LSB, 0},	 //YY in the XXYY address
  {CV_29_CONFIG, 128 },	 //Make sure this is 0 or else it will be random based on what is in the eeprom which could caue headaches
  {CV_VSTART, 16},
  {CV_ACCRATIO, 64},
  {CV_DECCRATIO, 64},
};

void(* resetFunc) (void) = 0;  //declare reset function at address 0


uint8_t FactoryDefaultCVIndex = sizeof(FactoryDefaultCVs) / sizeof(CVPair);
byte readSoundRom(uint8_t inNo, uint16_t inOffset);
uint16_t limitSpeed(uint16_t inSpeed);
void Motor_Control();


/*  *************************************************
    
    関数
    
    ************************************************* */

void notifyCVResetFactoryDefault()
{
	//When anything is writen to CV8 reset to defaults. 

	resetCVToDefault();	 
	Serial.println("Resetting...");
	delay(1000);  //typical CV programming sends the same command multiple times - specially since we dont ACK. so ignore them by delaying

	resetFunc();
};


void resetCVToDefault()
{
	//CVをデフォルトにリセット
	Serial.println("CVs being reset to factory defaults");
	
	for (int j=0; j < FactoryDefaultCVIndex; j++ ){
		Dcc.setCV( FactoryDefaultCVs[j].CV, FactoryDefaultCVs[j].Value);
	}
};

void notifyCVAck(void)
{
	Serial.println("notifyCVAck");

	if( gCVAckDir == 0)
	{
		
		analogWrite(MOTOR_PWM_B, 0);
		analogWrite(MOTOR_PWM_A, MOTOR_CVACK_POWER);
		delay( 6 );  
		analogWrite(MOTOR_PWM_A, 0);
		gCVAckDir = 128;
	}
	else
	{
		analogWrite(MOTOR_PWM_A, 0);
		analogWrite(MOTOR_PWM_B, MOTOR_CVACK_POWER);
		delay( 6 );  
		analogWrite(MOTOR_PWM_B, 0);
		gCVAckDir = 0;
	}


}

extern void	   notifyCVChange( uint16_t CV, uint8_t Value){
   //CVが変更されたときのメッセージ
   Serial.print("CV "); 
   Serial.print(CV); 
   Serial.print(" Changed to "); 
   Serial.println(Value, DEC);
};


void setup()
{
	uint8_t cv_value;
	
	//シリアル通信開始
	Serial.begin(115200);

	//D9,D10 PWM キャリア周期:31kHz
	TCCR1B &= B11111000;
	TCCR1B |= B00000001;
	
	//PWMを10bit化
	//TCCR1A |= B00000011;
	
	pinMode(ledPin, OUTPUT);
	digitalWrite(ledPin, HIGH);

	
	//PWM出力ピン D9,D10を出力にセット
	pinMode(9, OUTPUT);
	pinMode(10, OUTPUT);

	
	//DCCの応答用負荷ピン
	
	#if defined(DCCACKPIN)
	//Setup ACK Pin
	pinMode(DccAckPin,OUTPUT);
	digitalWrite(DccAckPin, 0);
	#endif
  
 

  #if !defined(DECODER_DONT_DEFAULT_CV_ON_POWERUP)
	if ( Dcc.getCV(CV_MULTIFUNCTION_PRIMARY_ADDRESS) == 0xFF ){	 //if eeprom has 0xFF then assume it needs to be programmed
	  Serial.println("CV Defaulting due to blank eeprom");
	  notifyCVResetFactoryDefault();
	  
   } else{
	 Serial.println("CV Not Defaulting");
   }
  #else
	 Serial.println("CV Defaulting Always On Powerup");
	 notifyCVResetFactoryDefault();
  #endif 
  
   
   
  // Setup which External Interrupt, the Pin it's associated with that we're using, disable pullup.
  Dcc.pin(0, 2, 0);
  
  // Call the main DCC Init function to enable the DCC Receiver
  Dcc.init( MAN_ID_DIY, 100,   FLAGS_MY_ADDRESS_ONLY , 0 ); 

  //Reset task
  gPreviousL5 = millis();
  
  //Init CVs
  /*
  gCV1_SAddr = Dcc.getCV( CV_MULTIFUNCTION_PRIMARY_ADDRESS ) ;
  gCVx_LAddr = (Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_MSB ) << 8) + Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_LSB );
  
  Serial.print("CV1(ShortAddr): ");
  Serial.println(gCV1_SAddr);
  Serial.print("CV17/18(LongAddr): ");
  Serial.println(gCVx_LAddr);
 
*/  Serial.println("Ready");
 
	 
}

void loop(){
	
	// You MUST call the NmraDcc.process() method frequently from the Arduino loop() function for correct library operation
	Dcc.process();
	

	if( (millis() - gPreviousL5) >= 100)
	{
		//Headlight control
		Motor_Control();
		
		//Reset task
		gPreviousL5 = millis();
		
	}
}

/* Motor control Task (10Hz) */
void Motor_Control()
{
	uint16_t aSpeedRef = gSpeedCmd;
	uint16_t aPWMRef = 0;
	uint16_t aLPF_Ratio = gCV3_AccRatio;
	uint16_t aSpeedRefLinear = 0;
	uint32_t aSpeedRef_offseted = 0;
	
	//Motor 出力
	
	//8bit LPF
	gPwmLPF_buf = ((gPwmLPF_buf * (256 - aLPF_Ratio)) >> 8) + ( aLPF_Ratio * aSpeedRef);

	//Linear Adjustment
	if( (gPwmLPF_buf >> 8) > 0)
	{
		//Accerelation
		aSpeedRef_offseted = (((MOTOR_LIM_MAX - (uint32_t)gCV2_Vstart) * (uint32_t)gPwmLPF_buf) >> 8) + ((uint32_t)gCV2_Vstart << 8);

		/* 16bitPWMでのリミッタ(後段は8bit) */
		if( (aSpeedRef_offseted >> 8) > MOTOR_LIM_MAX)
		{
		  aSpeedRef_offseted = MOTOR_LIM_MAX << 8;
		}
		
		aSpeedRefLinear = (uint16_t)aSpeedRef_offseted;
	}
	else
	{
		aSpeedRefLinear = 0;
	}

	gPrevSpeed = gPwmLPF_buf;

	//Limiter (8bit PWM Output)
	aPWMRef = limitSpeed((uint16_t)(aSpeedRefLinear >> 8));
	
	//PWM出力
	if( aPWMRef == 0)
	{
		analogWrite(MOTOR_PWM_A, 0);
		analogWrite(MOTOR_PWM_B, 0);
	}
	else
	{
		//進行方向でPWMのABを切り替える
		if( gPwmDir > 0){
			analogWrite(MOTOR_PWM_B, 0);
			analogWrite(MOTOR_PWM_A, aPWMRef);
		}
		else
		{
			analogWrite(MOTOR_PWM_A, 0);
			analogWrite(MOTOR_PWM_B, aPWMRef);
		}
	}

  
  //再生が終わっていたらここで関数を完了する
  switch (gSound_played)
  {
    case 1:
    case 11:
    case 3:
    case 31:
    {
      //draft sound,stop_sound時
      //連続のため終了しない
      break;
    }     
    case 2:
    case 4:
    {
      //kiteki_short,blast_soundの時
      if ((gState_F1 > 0) || (gState_F3 > 0))
      {
        //draft_sound,stop_soundに戻る
        break;
      }
      else
      {
        //再生終了
        return;
      }
    }
    default:
    {
      //0の時（音の鳴り始め）は下の分岐へ
      break;
    }
  }

  if( (gState_F2 > 0) && (gSound_playnext == 2))
  {
    //kiteki short
    gSound_played = gSound_playnext;
    gSound_playnext = 0;
    sample_rate = 8000;
    delay_offset_time = 0;
    repeat_sound = 1;
  }
  else if( (gState_F4 > 0) && (gSound_playnext == 4))
  {
    //blast_sound
    gSound_played = gSound_playnext;
    gSound_playnext = 0;
    sample_rate = 8000;
    delay_offset_time = 0;
    repeat_sound = 4;      
  }
  else if( (gState_F3 > 0))
  {
    //stop_sound
    if(gSound_playnext == 3)
    {
      gSound_playnext = 31;
    }
    else
    {
      gSound_playnext = 3;
    }            
    gSound_played = gSound_playnext;
    sample_rate = 8000;
    delay_offset_time = 250;
    repeat_sound = 1;      
  }
  //else if(( gState_F1 > 0) && ((gSound_playnext == 1) || (gSound_playnext == 11)))
  else if(gState_F1 > 0)
  {
    //draft sound
    //Speed0,1の時は、音は鳴らさない
    if(gSpeedRef <= 1)
    {
      return;
    }
    else
    {  
      //鳴り始めは5000Hz、最高速（128)で11000Hzとする。
      //125us  * 128 = 
      //990ms  - 8ms - 8ms - 50ms = 924 まで短くすればよい
      //sample_rate =  5000 + 6000 * aPWMRef / 256;
      sample_rate =  5000 + ((375 * aPWMRef) >> 4);
      //サウンドトグル,長さ取得
      if(gSound_playnext == 1)
      {
        gSound_playnext = 11;
      }
      else
      {
        gSound_playnext = 1;
      }            
      gSound_played = gSound_playnext;
      delay_offset_time = 100;
      repeat_sound = 1;
    }
  }
  //実際に鳴らすところ
  if (gSound_played != 0)
  {
    gSound_Length = getSoundLength(gSound_played);
    firstSample = readSoundRom(gSound_played, 0);
    lastSample = readSoundRom(gSound_played, gSound_Length - 1);
    float sample_time = (float)((float)gSound_Length * (float)repeat_sound + (float)firstSample + (float)lastSample + (float)delay_offset_time) /(float)8000 * (float)1000 * (float)8000 / (float)sample_rate;
    startPlayback();
    delay((unsigned long)sample_time);
    stopPlayback();
  }
}

uint16_t limitSpeed(uint16_t inSpeed)
{
	uint16_t aSpeed_ref = 0;
	
	if( inSpeed >= MOTOR_LIM_MAX)
	{
		aSpeed_ref = MOTOR_LIM_MAX;
	}
	else if( inSpeed <= 0)
	{
		aSpeed_ref = 0;
	}
/*	else if( (inSpeed > 0) && (inSpeed <= gCV2_Vstart))
	{
		aSpeed_ref = gCV2_Vstart;
	}*/
	else
	{
		aSpeed_ref = inSpeed;
	}
	
	return aSpeed_ref;
	
}


//DCC速度信号の受信によるイベント
extern void notifyDccSpeed( uint16_t Addr, uint8_t Speed, uint8_t ForwardDir, uint8_t MaxSpeed )
{
	
	uint16_t aSpeedRef = 0;
	
	//速度値の正規化(255を100%とする処理)
	if( Speed >= 1)
	{
		aSpeedRef = ((Speed - 1) * 255) / MaxSpeed;
	}
	else
	{
		aSpeedRef = 0;
	}
	
	//リミッタ
	if(aSpeedRef > 255)
	{
		aSpeedRef = 255;
	}
	
	gSpeedCmd = aSpeedRef;
	gPwmDir = ForwardDir;
	
	gSpeedRef = Speed;
	
	// デバッグメッセージ
/*	Serial.print("Speed - ADR: ");
	Serial.print(Addr);
	Serial.print(", SPD: ");
	Serial.print(Speed);
	Serial.print(", DIR: ");
	Serial.print(ForwardDir);
	Serial.print(", MAX: ");
	Serial.println(MaxSpeed);
*/
}

//ファンクション信号受信のイベント
extern void notifyDccFunc( uint16_t Addr, FN_GROUP FuncGrp, uint8_t FuncState)
{
	
	if( FuncGrp == FN_0_4)
	{
		if( gState_F0 != (FuncState & FN_BIT_00))
		{
			//Get Function 0 (FL) state
			gState_F0 = (FuncState & FN_BIT_00);
		}
		if( gState_F1 != (FuncState & FN_BIT_01))
		{
			gSound_played = 0;
			gSound_playnext = 1;
			//Get Function 1 state
			gState_F1 = (FuncState & FN_BIT_01);
		}
		if( gState_F2 != (FuncState & FN_BIT_02))
		{
			gSound_played = 0;
			gSound_playnext = 2;
			//Get Function 2 state
			gState_F2 = (FuncState & FN_BIT_02);
		}
		if( gState_F3 != (FuncState & FN_BIT_03))
		{
			gSound_played = 0;
			gSound_playnext = 3;
			//Get Function 3 state
			gState_F3 = (FuncState & FN_BIT_03);
		}
		if( gState_F4 != (FuncState & FN_BIT_04))
		{
			gSound_played = 0;
			gSound_playnext = 4;
			//Get Function 4 state
			gState_F4 = (FuncState & FN_BIT_04);
		}
	}

}

byte readSoundRom(uint8_t inNo, uint16_t inOffset)
{
  //SoundROMの生データ取得
	switch(inNo)
	{
	case 1:
		return pgm_read_byte(&draft_sound1_100_data[inOffset]);
		break;
	case 11:
		return pgm_read_byte(&draft_sound2_100_data[inOffset]);
		break;
	case 2:
		return pgm_read_byte(&kiteki_short_data[inOffset]);
		break;
	case 3:
		return pgm_read_byte(&stop_sound1_150_data[inOffset]);
		break;		
	case 31:
		return pgm_read_byte(&stop_sound2_150_data[inOffset]);
		break;		
	case 4:
		return pgm_read_byte(&blast_sound_data[inOffset]);
		break;		
	default:
		return 0;
		break;
	}
}

int getSoundLength(uint8_t inNo)
{
  //sound length の取得
	switch(inNo)
	{
	case 1:
		return draft_sound1_100_length;
		break;
	case 11:
		return draft_sound2_100_length;
		break;
	case 2:
		return kiteki_short_length;
		break;
	case 3:
		return stop_sound1_150_length;
		break;		
	case 31:
		return stop_sound2_150_length;
		break;		
	case 4:
		return blast_sound_length;
		break;		
	default:
		return 0;
		break;
	}
}  
void stopPlayback()
{
    // Disable playback per-sample interrupt.
    TIMSK1 &= ~_BV(OCIE1A);

    // Disable the per-sample timer completely.
    TCCR1B &= ~_BV(CS10);

    // Disable the PWM timer.
    TCCR2B &= ~_BV(CS10);

    digitalWrite(speakerPin, LOW);
}

// This is called at 8000 Hz to load the next sample.
ISR(TIMER1_COMPA_vect)
{
  switch(s_state)
  {
    //fade in
    case s_start:
    {
      if(sample >= firstSample)
      {
        s_state = s_sound;
        sample = 0;
        break;
      }
      else
      {
        if(speakerPin==11)
        {
            OCR2A = sample;
        }
        else
        {
            OCR2B = sample;                
        }
      }
      sample ++;
      break;
    }
    case s_sound:
    {
      //play sound
      if(sample == gSound_Length - 1)
      {
        sample = 0;
        //repeat有り無し
        if(repeat_sound == 1)
        {
          s_state = s_end;
          break;
        }
        else
        {
          repeat_sound--;
        }
      }
      else
      {
        if(speakerPin==11)
        {
            OCR2A = readSoundRom(gSound_played, sample);
        }
        else
        {
            OCR2B = readSoundRom(gSound_played, sample);            
        }
      }
      sample++;
      break;
    }
    case s_end:
    {
      //fade out
      if(sample >= lastSample)
      {
        s_state = s_idle;
        sample = 0;
      }
      else
      {
        if(speakerPin==11)
        {
            OCR2A = lastSample - sample;
        }
        else
        {
            OCR2B = lastSample - sample;
        }
      }
      sample ++;
      break;
    }
    default:
    {
      s_state = s_idle;
    }
  }
}

void startPlayback()
{
	
    pinMode(speakerPin, OUTPUT);

    //State 
    s_state = s_start;
    // Set up Timer 2 to do pulse width modulation on the speaker
    // pin.

    // Use internal clock (datasheet p.160)
    ASSR &= ~(_BV(EXCLK) | _BV(AS2));

    // Set fast PWM mode  (p.157)
    TCCR2A |= _BV(WGM21) | _BV(WGM20);
    TCCR2B &= ~_BV(WGM22);

    if(speakerPin==11){
        // Do non-inverting PWM on pin OC2A (p.155)
        // On the Arduino this is pin 11.
        TCCR2A = (TCCR2A | _BV(COM2A1)) & ~_BV(COM2A0);
        TCCR2A &= ~(_BV(COM2B1) | _BV(COM2B0));
        // No prescaler (p.158)
        TCCR2B = (TCCR2B & ~(_BV(CS12) | _BV(CS11))) | _BV(CS10);

        // Set initial pulse width to the first sample.
        //OCR2A = readSoundRom(gSound_played, 0);
        OCR2A = 0;
    } else {
        // Do non-inverting PWM on pin OC2B (p.155)
        // On the Arduino this is pin 3.
        TCCR2A = (TCCR2A | _BV(COM2B1)) & ~_BV(COM2B0);
        TCCR2A &= ~(_BV(COM2A1) | _BV(COM2A0));
        // No prescaler (p.158)
        TCCR2B = (TCCR2B & ~(_BV(CS12) | _BV(CS11))) | _BV(CS10);

        // Set initial pulse width to the first sample.
        //OCR2B = readSoundRom(gSound_played, 0);
        OCR2B = 0;
    }





    // Set up Timer 1 to send a sample every interrupt.

    cli();

    // Set CTC mode (Clear Timer on Compare Match) (p.133)
    // Have to set OCR1A *after*, otherwise it gets reset to 0!
    TCCR1B = (TCCR1B & ~_BV(WGM13)) | _BV(WGM12);
    TCCR1A = TCCR1A & ~(_BV(WGM11) | _BV(WGM10));

    // No prescaler (p.134)
    TCCR1B = (TCCR1B & ~(_BV(CS12) | _BV(CS11))) | _BV(CS10);

    // Set the compare register (OCR1A).
    // OCR1A is a 16-bit register, so we have to do this with
    // interrupts disabled to be safe.
    //OCR1A = F_CPU / SAMPLE_RATE;    // 16e6 / 8000 = 2000
    OCR1A = F_CPU / sample_rate;    // 16e6 / 8000 = 2000
    // Enable interrupt when TCNT1 == OCR1A (p.136)
    TIMSK1 |= _BV(OCIE1A);


    sample = 0;
    sei();
}
